% Copyright (c) [2024] [Daniele Massaro, FLOW, Engineering Mechanics, KTH Royal Institute of Technology, 2024 (dmassaro@kth.se)]
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
%
% 1. Redistributions of source code must retain the above copyright
%    notice, this list of conditions and the following disclaimer.
%
% 2. Redistributions in binary form must reproduce the above copyright
%    notice, this list of conditions and the following disclaimer in the
%    documentation and/or other materials provided with the distribution.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
% IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
% TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
% PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
% HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
% SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
% LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
% DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
% THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
% OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

% -------------------------------------------------------------------------
% The code computes the Voronoi areas for a set of points (x,y), 
% providing a finite Voronoi area to the boundary points as well
% -------------------------------------------------------------------------

clear all
close all
clc

% Initial set of points. 
% For example, they correspond to (y_p/R,l_y/R) in Figure 4a in the manuscript. 
x = [0.086 0.2625 0.801 0.0292 0.9289 0.7303 0.4886 0.5785 0.2373 0.4588]'; 
y = [0.9631 0.5468 0.5211 0.2316 0.4889 0.6241 0.5791 0.3955 0.3674 0.98]';


% Native MATLAB alogorithm for Voronoi tesselation (at the boundary
% points an infinity area is generated)
figure(1)
voronoi(x,y)
hold on
% The points are marked with black asterisks.
plot(x,y,'k*')
xlabel('x-points')
ylabel('y-points')
axis([-0.8 1.6 -0.3 1.5])

% Boundary points indicated with encircled asterisks and a line connecting these points
j = boundary(x,y);
hold on;
xb = x(j);
yb = y(j);
plot(xb,yb,'ko-');


% Expansion factor: A large area must be attributed to the boundary points.
% The factor A should be as large as possible (at least a factor of 2).
% A large area must be ascribed to these points since they are not surrounded
% by other points on one side (at least).
A = 2;
xb_e = A*xb+(1-A)*mean(xb(1:end-1));
yb_e = A*yb+(1-A)*mean(yb(1:end-1));
plot(xb_e,yb_e,'ro-');

% Augmenting the set of points with the extra points (xb_e,yb_e) (red circles)
X = [xb_e(1:end-1); x];
Y = [yb_e(1:end-1); y];

% Voronoi tesselation including the extra points (xb_e,yb_e)
figure(2)
voronoi(X,Y)
hold on
plot(X,Y,'k*')
plot(xb,yb,'ko-');
xlabel('x-points')
ylabel('y-points')
axis([-0.8 1.6 -0.3 1.5])
a =1;
 
% Voronoi vertices v and the Voronoi cells c of the Voronoi diagram 
[v,c] = voronoin([X Y]) ;


for i = 1:length(c)
    % To identify extra boundary points (xb_e,yb_e) with an infinity area
    v1 = v(c{i},1) ; 
    v2 = v(c{i},2) ;
    TF1 = isinf(v1);
    TF2 = isinf(v2);
    TF1 = int64(TF1);
    TF2 = int64(TF2);

    % To exclude extra boundary points (xb_e,yb_e) with an infinity area
    if all(TF1==0) && all(TF2==0) 
        % To determine whether a point is inside a polygon 
        in = inpolygon(x,y,v1,v2);    
        IN =int64(in);
        for z = 1:length(IN)
            % Selecting a specific point from the initial set that is 
            % inside a given polygon with vertices defined by v.
            if IN(z)==1
                % Voronoi area calculation for the original set of points
                areas(z)= polyarea(v1,v2);
            end
        end
    end
end
fprintf('The values of the Voronoi areas for the initial set of points: \n %s\n', mat2str(areas));

